/*
**  NGPT - Next Generation POSIX Threading
**  Copyright (c) 2001 IBM Corporation <babt@us.ibm.com>
**  Portions Copyright (c) 1999-2000 Ralf S. Engelschall <rse@engelschall.com>
**
**  This file is part of NGPT, a non-preemptive thread scheduling
**  library which can be found at http://www.ibm.com/developer.
**
**  This library is free software; you can redistribute it and/or
**  modify it under the terms of the GNU Lesser General Public
**  License as published by the Free Software Foundation; either
**  version 2.1 of the License, or (at your option) any later version.
**
**  This library is distributed in the hope that it will be useful,
**  but WITHOUT ANY WARRANTY; without even the implied warranty of
**  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
**  Lesser General Public License for more details.
**
**  You should have received a copy of the GNU Lesser General Public
**  License along with this library; if not, write to the Free Software
**  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
**  USA.
**
**  pth_cancel.c: Pth thread cancellation
*/
                             /* ``Study it forever and you'll still wonder.
                                  Fly it once and you'll know.''
                                                       -- Henry Spencer */
#include "pth_p.h"

/* set cancellation state */
void pth_cancel_state(int newstate, int *oldstate)
{
    pth_t current = pth_get_current();
    if (oldstate != NULL)
        *oldstate = current->cancelstate;
    if (newstate != 0)
        current->cancelstate = newstate;
    return;
}

/* enter a cancellation point */
void pth_cancel_point(void)
{
    pth_t current = pth_get_current();
    if (   current->cancelreq == TRUE
        && current->cancelstate & PTH_CANCEL_ENABLE) {
        /* avoid looping if cleanup handlers contain cancellation points */
        current->cancelreq = FALSE;
        pth_debug2("pth_cancel_point: terminating cancelled thread \"%s\"", current->name);
        pth_exit(PTH_CANCELED, FALSE);
    }
    return;
}

/* cancel a thread (the friendly way) */
int pth_cancel(pth_t thread)
{
    pth_pqueue_t *q;
    pth_descr_t descr = pth_get_native_descr();
    pid_t tid = descr->tid;
    pth_t current = descr->current;

    if (thread == NULL || !pth_thread_exists(thread))
        return FALSE;

    /* the thread has to be at least still alive */
    if (thread->state == PTH_STATE_DEAD || thread->cancelreq == TRUE)
	return TRUE;

    /* now mark the thread as cancelled */
    thread->cancelreq = TRUE;

    /* when cancellation is enabled in async mode we cancel the thread immediately */
    if (   thread->cancelstate & PTH_CANCEL_ENABLE
        && thread->cancelstate & PTH_CANCEL_ASYNCHRONOUS) {

	if (thread == current) {
	    pth_cancel_point();
	    return TRUE;
	}

        /* remove thread from its queue */
        switch (thread->state) {
            case PTH_STATE_NEW:     q = &pth_NQ; break;
            case PTH_STATE_READY:   return TRUE;
            case PTH_STATE_WAITING: return TRUE;
            case PTH_STATE_MUTEX_WAITING: return TRUE;
            default:                q = NULL;
        }
        if (q == NULL)
            return FALSE;

	if (!pth_pqueue_contains(q, thread))
                return FALSE;
	_pth_acquire_lock(&(q->q_lock), tid);
	pth_pqueue_delete(q, thread);
	_pth_release_lock(&(q->q_lock), tid);
    
        /* execute cleanups */
        pth_thread_cleanup(thread);

        /* and now either kick it out or move it to dead queue */
        if (!thread->joinable) {
            pth_debug2("pth_cancel: kicking out cancelled thread \"%s\" immediately", thread->name);
            pth_tcb_free(thread);
        }
        else {
            pth_debug2("pth_cancel: moving cancelled thread \"%s\" to dead queue", thread->name);
            thread->join_arg = PTH_CANCELED;
            thread->state = PTH_STATE_DEAD;
	    _pth_acquire_lock(&(pth_DQ.q_lock), tid);
            pth_pqueue_insert(&pth_DQ, PTH_PRIO_STD, thread);
	    _pth_release_lock(&(pth_DQ.q_lock), tid);
        }
    }
    return TRUE;
}

/* abort a thread (the cruel way) */
int pth_abort(pth_t thread)
{
    if (thread == NULL)
        return FALSE;

    /* the current thread cannot be aborted */
    if (thread == pth_get_current())
        return FALSE;

    if (thread->state == PTH_STATE_DEAD && thread->joinable) {
        /* if thread is already terminated, just join it */
        if (!pth_join(thread, NULL))
            return FALSE;
    }
    else {
        /* else force it to be detached and cancel it asynchronously */
        thread->joinable = FALSE;
        thread->cancelstate = (PTH_CANCEL_ENABLE|PTH_CANCEL_ASYNCHRONOUS);
        if (!pth_cancel(thread))
            return FALSE;
    }
    return TRUE;
}

